# Copyright 2022 Lawrence Livermore National Security, LLC and other
# HPCIC DevTools Developers. See the top-level COPYRIGHT file for details.
#
# SPDX-License-Identifier: (MIT)

from ..base import Prices


class GoogleCloudPrices(Prices):
    """
    Google Compute Engine prices
    """

    pass
